/*
 * LED_cube_test.c
 *
 * Created: 17.07.2013 11:16:27
 *  Author: Florian
 *-U lfuse:w:0xa4:m -U hfuse:w:0xd9:m
 */ 

//----------------------------------------------------------------------
// System includes
//----------------------------------------------------------------------
#include <avr/io.h>
#define F_CPU 8000000UL  // 8 MHZ
#define __DELAY_BACKWARD_COMPATIBLE__
#include <util/delay.h>
#include <stdlib.h>
#include <avr/interrupt.h>


//----------------------------------------------------------------------
// Declarations
//----------------------------------------------------------------------

//Port B
#define E0   0b00000001 //PB0
#define E1   0b00000010 //PB1
#define E2   0b00000100 //PB2
#define E3   0b00001000 //PB3
#define L_32 0b01000000 //PB6
#define L_33 0b10000000 //PB7
//Port C
#define L_20 0b00000001 //PC0
#define L_21 0b00000010 //PC1
#define L_22 0b00000100 //PC2
#define L_23 0b00001000 //PC3
#define L_30 0b00010000 //PC4
#define L_31 0b00100000 //PC5
//Port D
#define L_00 0b00000001 //PD0
#define L_01 0b00000010 //PD1
#define L_02 0b00000100 //PD2
#define L_03 0b00001000 //PD3
#define L_10 0b00010000 //PD4
#define L_11 0b00100000 //PD5
#define L_12 0b01000000 //PD6
#define L_13 0b10000000 //PD7
//Constants 
#define fluid_delay 175	//Time which is needed that the animation looks fluid
volatile uint8_t voxel[4][4][4];
volatile int currentLayer;
const char copyright [] = "Copyright (c) Florian Brderlin / Nur fuer privaten Gebrauch / For private use only";


//----------------------------------------------------------------------
//Function declarations 
//----------------------------------------------------------------------
void init(void);			//Initialization of I/O - Ports
void setLayer (int lyr);	//Push the layer lyr to GND
void clearLayer ();			//Turn out all layers
void setXY (int x, int y);	//Calculate the right LED form x and y
void setLED (int idx);		//Pull the LED with idx to VCC 
void clearLED (void);		//Turn all LED out
void checkLED (void);		//Checking all LEDs after booting
void rainEffect(int iterations);	//Let it rain!	
void setVoxelXYZ (int x, int y, int z, int value);
void setAllVoxels (int value);
void setVoxelX(int x, int value);
void setVoxelY(int y, int value);
void setVoxelZ(int z, int value);
void followEffect(int iterations);
void popUpEffect(int iterations);
void pulsationEffect(int iterations, ...);




//----------------------------------------------------------------------
//Main function
//----------------------------------------------------------------------
int main(void)
{
	init();
	
	setAllVoxels(0);
	
	// Enable interrupts to start drawing the cube buffer.
	// When interrupts are enabled, ISR(TIMER2_COMP_vect)
	// will run on timed intervals.
	sei();
	
	checkLED();
	_delay_ms(1000);
    while(1)
    {		_delay_ms(1000);
		rainEffect(100);
		_delay_ms(2000);
		followEffect(20);
		_delay_ms(2000);
		popUpEffect(70);
    }
}


//----------------------------------------------------------------------
//Initialization 
//----------------------------------------------------------------------
void init()
{
	//Initialization of I/O - Ports
	DDRB = (L_32|L_33|E0|E1|E2|E3);						//Port B 0-3 und 6-7 to Output; 4+5 MISO / SCK
	DDRC = (L_20|L_21|L_22|L_23|L_30|L_31);				//Port C 0-5 to Output; 6 RSET; 7 doesn't exist
	DDRD = (L_00|L_01|L_02|L_03|L_10|L_11|L_12|L_13);	//Port D 0-7 (all) to Output 
	
	//Initialization of Interrupt Timer2
	TCNT2 = 0x00;	// initial counter value = 0;
	TIMSK |= (1 << OCIE2); // Enable CTC interrupt
	
	//Interrupt Frame rate
	//8 MHz = 8,000,000 counts /sec
	//8,000,000/1024/10/4 = 195 FPS
	//There are 4 layers to update..
	OCR2 = 10; 				// interrupt at counter = 10
	TCCR2 = 0x07; 			// prescaler = 1024
	TCCR2 |= (1 << WGM21);	// Clear after Interrupt Mode to 0
	
	currentLayer = 0;
	
}

ISR	(TIMER2_COMP_vect)
{
	clearLayer();
	clearLED();
	int x;
	for (x = 0; x < 4; x++)
	{
		int y;
		for (y = 0; y < 4; y++)
		{
			if (voxel[x][y][currentLayer])
			{
				setXY(x, y);
			}	
		}
	}
			
	setLayer(currentLayer);
	
	if (currentLayer++ == 3)
	{
		currentLayer = 0;
	}
}

void setLayer (int lyr)
{
	switch (lyr)
	{
		case 0: PORTB |= E0; break;
		case 1: PORTB |= E1; break;
		case 2: PORTB |= E2; break;
		case 3: PORTB |= E3; break;
	}
}

void clearLayer ()
{
	PORTB &= ~(E0 | E1 | E2 | E3); 
}

void setVoxelX(int x, int value)
{
	int y,z;
	for (y = 0; y < 4; y++)
	{
		for (z = 0; z < 4; z++)
		{
			setVoxelXYZ(x,y,z,value);
		}
	}
}

void setVoxelY(int y, int value)
{
	int x,z;
	for (x = 0; x < 4; x++)
	{
		for (z = 0; z < 4; z++)
		{
			setVoxelXYZ(x,y,z,value);
		}
	}
}

void setVoxelZ(int z, int value)
{
	int x,y;
	for (x = 0; x < 4; x++)
	{
		for (y = 0; y < 4; y++)
		{
			setVoxelXYZ(x,y,z,value);
		}
	}
}

void setAllVoxels (int value)
{
	int z;
	for (z = 0; z < 4; z++)
	{
		setVoxelZ(z,value);
	}
	
}

void setVoxelXYZ (int x, int y, int z, int value)
{
	voxel[x][y][z] = value;
}

void setLED (int idx)
{
	switch (idx)
	{
		case  0: PORTD |= L_00; break;
		case  1: PORTD |= L_01; break;
		case  2: PORTD |= L_02; break;
		case  3: PORTD |= L_03; break;
		case  4: PORTD |= L_10; break;
		case  5: PORTD |= L_11; break;
		case  6: PORTD |= L_12; break;
		case  7: PORTD |= L_13; break;
		case  8: PORTC |= L_20; break;
		case  9: PORTC |= L_21; break;
		case 10: PORTC |= L_22; break;
		case 11: PORTC |= L_23; break;
		case 12: PORTC |= L_30; break;
		case 13: PORTC |= L_31; break;
		case 14: PORTB |= L_32; break;
		case 15: PORTB |= L_33; break;	
	}
}

void setXY (int x, int y)
{
	setLED(y*4+x);
}


void clearLED ()
{
	PORTB &= ~(L_32|L_33);
	PORTC &= ~(L_20|L_21|L_22|L_23|L_30|L_31);
	PORTD &= ~(L_00|L_01|L_02|L_03|L_10|L_11|L_12|L_13);
}


void checkLED()
{
	int x,y,z;
	for (x = 0; x < 4; x++)
	{
		setVoxelX(x,1);
		_delay_ms(fluid_delay);
		setVoxelX(x,0);
	}
	for (y = 0; y < 4; y++)
	{
		setVoxelY(y,1);
		_delay_ms(fluid_delay);
		setVoxelY(y,0);
	}
	for (z = 0; z < 4; z++)
	{
		setVoxelZ(z,1);
		_delay_ms(fluid_delay);
		setVoxelZ(z,0);
	}
	
}		


void rainEffect(int iterations)
{
	int i,x,y;
	int lyr; //following layer (downwards)
	
	for (i = 0; i < iterations; i++)
	{
		//Clear 3rd layer
		setVoxelZ(3,0);
		int rand,r;
		//guess amount of raindrops
		rand = 1;//((random()%2)+1);
		//Setting random amount of random pixel on layer 3
		for (r = 0; r < rand; r++)
		{
			setVoxelXYZ(random()%4,random()%4,3,1);
		}
		_delay_ms(fluid_delay);
		
		//Raindrop has to fall
		//from layer 3 to 0 and then disappear
		//!!!for-loop ends by lyr = 2, because layer 3 will be created new
		for (lyr = 0; lyr < 3; lyr++)
		{
			for (x = 0; x < 4; x++)
			{
				for (y = 0; y < 4; y++)
				{
					setVoxelXYZ(x,y,lyr,voxel[x][y][lyr+1]);
				}
			}
		}
	}
	setAllVoxels(0);
}


void followEffect(int interations)
{
	int i,j,x,y,z;
	char graphicSaver [4][4][4];
	for (i = 0; i < 4; i++)
	{
		//drawing two stairways
		setVoxelXYZ(i,0,i,1);
		setVoxelXYZ(3-i,3,i,1);
	}
	for (j = 0; j < interations; j++)
	{
		//It takes 12 steps to go one time around
		for (i = 0; i < 12; i++)
		{
			for (x = 0; x < 4; x++)
			{
				for (y = 0; y < 4; y++)
				{
					for (z = 0; z < 4; z++)
					{
						graphicSaver[x][y][z] = voxel[x][y][z];
					}
				}	
			}
			for (x = 1; x < 4; x++)
			{
				for (z = 0; z < 4; z++)
				{
					setVoxelXYZ(x,0,z,graphicSaver[x-1][0][z]);
				}
			}
			for (y = 1; y < 4; y++)
			{
				for (z = 0; z < 4; z++)
				{
					setVoxelXYZ(3,y,z,graphicSaver[3][y-1][z]);
				}
			}
			for (x = 2; x >= 0; x--)
			{
				for (z = 0; z < 4; z++)
				{
					setVoxelXYZ(x,3,z,graphicSaver[x+1][3][z]);
				}
			}
			for (y = 2; y >= 0; y--)
			{
				for (z = 0; z < 4; z++)
				{
					setVoxelXYZ(0,y,z,graphicSaver[0][y+1][z]);
				}
			}
			_delay_ms(fluid_delay);
		}
	}
}



void popUpEffect(int iterations)
{
	char firstLayerRegister [4][4];	//These register save the current waiting position of each LED ([x][y])
	int i,j;
	
	setAllVoxels(0);	//Clear all
	setVoxelZ(0,1);		//Turn on all LEDs that are located in layer 0
	
	//At the beginning all LEDs are located in the 0. layer
	for (i = 0; i < 4; i++)
	{
		for (j = 0; j < 4; j++)
		{
			firstLayerRegister[i][j] = 1;
		}
	}
	
	for (i = 0; i < iterations; i++)
	{
		int x,y,z;
		//guess a random LED
		x = (random()%4);
		y = (random()%4);
		
		//to examine whether the LED is located in the first or in the third layer
		if (firstLayerRegister[x][y])
		{
			//swoop positions
			firstLayerRegister[x][y] = 0;
			//LED pop up
			for (z = 1; z < 4; z++)
			{
				setVoxelXYZ(x,y,z,1);
				setVoxelXYZ(x,y,z-1,0);
				_delay_ms(fluid_delay/z);
			}
		}
		else
		{
			//swoop positions
			firstLayerRegister[x][y] = 1;
			//LED pop up
			for (z = 2; z >= 0; z--)
			{
				setVoxelXYZ(x,y,z,1);
				setVoxelXYZ(x,y,z+1,0);
				_delay_ms((fluid_delay/3)*z);
			}
		}
		_delay_ms(1000);
		
	}
	
}